from finq import FINQ


def test_random_select():
    a = [1, 2, 5, 7, 12, 15, 22, 26]
    expected = [1, 2, 5, 7, 12, 15, 22, 26]

    assert FINQ(a).random(2).to_list() == expected


def test_random_sort():
    a = [1, 2]
    expected1 = [1, 2]
    expected2 = [2, 1]
    random_sort = FINQ(a).shuffle().to_list()
    assert random_sort == expected1 or random_sort == expected2



from finq import FINQ, First, RPairWith, PairWith, Second


def test_dict_builds():
    a = [1, 1, 2, 3, 3, 3]
    expected = {0: [2], 1: [1, 1, 3, 3, 3]}
    func = lambda i: i % 2

    assert FINQ(a).group_by(func).map(RPairWith(First, func)).to_dict() == expected


def test_dict_builds_with_custom_key():
    a = [1, 1, 2, 3, 3, 3]
    expected = {0: [2], 1: [1, 1, 3, 3, 3]}
    func = lambda i: i % 2

    assert FINQ(a).group_by(func).map(PairWith(First, func)).to_dict(Second, First) == expected



from finq import FINQ


def odd(a):
    return a % 2 == 1


def test_filters_odd():
    a = [1, 2, 5, 7, 12, 15, 22, 26]
    expected = [1, 5, 7, 15]

    a_f = FINQ(a)

    assert a_f.filter(odd).to_list() == expected


def test_filters_none():
    a = [1, 2, 5, 7, 12, 15, 22, 26]
    expected = []

    a_f = FINQ(a)

    assert a_f.filter(lambda a: False).to_list() == expected


def test_filters_nonbool():
    a = [1, 2, 5, 7, 12, 15, 22, 26]
    expected = [1, 2, 5, 7, 15, 22, 26]

    a_f = FINQ(a)
    assert a_f.filter(lambda a: a - 12).to_list() == expected



from finq import FINQ


def test_any():
    a = [1, 2, 5, 7]
    expected = [5, 7]

    a_f = FINQ(a)
    func = lambda i: i > 4

    assert a_f.any(func)
    assert a_f.filter(func).to_list() == expected


def test_any_false():
    a = [1, 2, 5, 7]
    expected = []

    a_f = FINQ(a)
    func = lambda i: i > 40

    assert not a_f.any(func)
    assert a_f.filter(func).to_list() == expected

def test_none_false():
    a = [1, 2, 5, 7]
    expected = [5, 7]

    a_f = FINQ(a)
    func = lambda i: i > 4

    assert not a_f.none(func)
    assert a_f.filter(func).to_list() == expected


def test_none():
    a = [1, 2, 5, 7]
    expected = []

    a_f = FINQ(a)
    func = lambda i: i > 40

    assert a_f.none(func)
    assert a_f.filter(func).to_list() == expected



from finq import FINQ


def test_flatten_once():
    a = [[1], 4, [5]]
    expected = [1, 4, 5]

    assert FINQ(a).flatten().to_list() == expected


def test_flatten_more_than_once():
    a = [[[[1]]], 4, [5]]
    expected = [1, 4, 5]

    assert FINQ(a).flatten().to_list() == expected


def test_doesnt_flatten_flat_collection():
    a = [1, 4, 5]
    expected = [1, 4, 5]

    assert FINQ(a).flatten().to_list() == expected



from finq import FINQ


def test_flat_map_list_of_lists_to_list():
    a = [[1, 2, 5], [7, 12], [15, 22, 26]]
    expected = [1, 2, 5, 7, 12, 15, 22, 26]

    a_f = FINQ(a)

    assert a_f.flat_map().to_list() == expected


def test_flat_map_list_of_tuples_to_list():
    a = [(1, 2, 5), (7, 12), (15, 22, 26)]
    expected = [1, 2, 5, 7, 12, 15, 22, 26]

    a_f = FINQ(a)

    assert a_f.flat_map().to_list() == expected


def test_flat_map_tuple_of_list_to_list():
    a = ([1, 2, 5], [7, 12], [15, 22, 26])
    expected = [1, 2, 5, 7, 12, 15, 22, 26]

    a_f = FINQ(a)

    assert a_f.flat_map().to_list() == expected


def test_flat_map_set_of_list_to_list():
    a = {(1, 2, 5), (7, 12), (15, 22, 26)}
    expected = [1, 2, 5, 7, 12, 15, 22, 26]

    a_f = FINQ(a)

    assert a_f.flat_map().to_list() == expected



from finq import FINQ


def test_skips():
    a = [1, 2, 5, 7, 12, 15, 22, 26]
    expected = [7, 12, 15, 22, 26]

    a_f = FINQ(a)

    assert a_f.skip(3).to_list() == expected


def test_takes():
    a = ['7', '15', '2', '22', '1', '12', '5', '26']
    expected = ['7', '15', '2', '22']

    a_f = FINQ(a)

    assert a_f.take(4).to_list() == expected


def test_takes_more_than_exist():
    a = [1, 2, 5, 7, 12, 15, 22, 26]
    expected = [1, 2, 5, 7, 12, 15, 22, 26]

    a_f = FINQ(a)

    assert a_f.take(40).to_list() == expected


def test_skips_more_than_exist():
    a = [1, 2, 5, 7, 12, 15, 22, 26]
    expected = []

    a_f = FINQ(a)

    assert a_f.skip(40).to_list() == expected



from finq import FINQ


def test_pairs():
    a = [1, 2, 5, 7]
    expected = [(0, 1), (1, 2), (2, 5), (3, 7)]

    a_f = FINQ(a)

    assert a_f.enumerate().to_list() == expected


def test_pairs_from_1():
    a = [1, 2, 5, 7]
    expected = [(1, 1), (2, 2), (3, 5), (4, 7)]

    a_f = FINQ(a)

    assert a_f.enumerate(1).to_list() == expected



from finq import FINQ


def test_count():
    a = [1, 2, 5, 7, 12, 15, 22, 26]
    expected = 8

    a_f = FINQ(a)

    assert a_f.count() == expected


def test_max():
    a = [1, 2, 5, 7, 12, 15, 22, 26]
    expected = 26

    a_f = FINQ(a)

    assert a_f.max() == expected


def test_min():
    a = [1, 2, 5, 7, 12, 15, 22, 26]
    expected = 1

    a_f = FINQ(a)

    assert a_f.min() == expected


def test_sum():
    a = [1, 2, 5, 7, 12, 15, 22, 26]
    expected = 90

    a_f = FINQ(a)

    assert a_f.sum() == expected


def test_max_diff():
    a = [1, 2, 5, 7, 12, 15, 22, 26]
    expected = 25

    a_f = FINQ(a)

    assert a_f.max_diff() == expected


def test_max_diff_with_generator():
    expected = 25

    a_f = FINQ(i for i in [1, 2, 5, 7, 12, 15, 22, 26])

    assert a_f.max_diff() == expected



from finq import FINQ


def collector(l: list):
    def write(s: str):
        l.append(s)

    write.collector = l

    return write


def test_foreach():
    a = [1, 2, 5, 7]
    expected = ['1', '2', '5', '7']

    mock = collector([])
    a_f = FINQ(a)

    a_f.map(str).for_each(mock)

    assert mock.collector == expected



from finq import Identity, Consumer, IdentityTrue, IdentityFalse, Sum, Multiply, Square, PairSum, TupleSum, First, \
    Second, FINQ, PairWith, Compose


def test_identity():
    a = 125123621
    assert Identity(a) == a


def test_consumer():
    a = 125123621
    assert Consumer(a) is None


def test_identity_true():
    a = 125123621
    assert IdentityTrue(a)


def test_identity_false():
    a = 125123621
    assert not IdentityFalse(a)


def test_sum():
    a = 125123621
    b = 125123622
    assert Sum(a, b) == a + b


def test_multiply():
    a = 125123621
    b = 125123622
    assert Multiply(a, b) == a * b


def test_square():
    a = 125123621
    assert Square(a) == a * a


def test_pair_sum():
    a = 125123621, 125123622
    assert PairSum(a) == a[0] + a[1]


def test_pair_tuple_sum():
    a = 125123621, 125123622
    assert TupleSum(a) == a[0] + a[1]


def test_empty_tuple_sum():
    a = ()
    assert TupleSum(a) == 0


def test_singular_tuple_sum():
    a = (125123621,)
    assert TupleSum(a) == a[0]


def test_triple_tuple_sum():
    a = (125123621, 125123621, 125123621)
    assert TupleSum(a) == a[0] * 3


def test_first():
    a = 125123621, 125123622
    assert First(a) == a[0]


def test_second():
    a = 125123621, 125123622
    assert Second(a) == a[1]


def test_pair_with_identity():
    a = [1, 1, 2, 3, 3, 3]
    expected = [(1, 1), (1, 1), (2, 2), (3, 3), (3, 3), (3, 3)]

    assert FINQ(a).map(PairWith(Identity)).to_list() == expected


def a(w):
    def _a(x):
        w.append('a' + str(x))
        return x + 1

    return _a


def b(w):
    def _b(x):
        w.append('b' + str(x))
        return x - 1

    return _b


def test_compose():
    ll = []

    expectedF = Identity
    composition = Compose(a(ll), b(ll))

    expected = FINQ(range(100))
    composition(0)
    assert ll == ['a0', 'b1']
    assert FINQ(range(100)).map(composition).to_list() == expected.map(expectedF).to_list()



from random import random
from typing import Tuple, Callable, NoReturn

from finq.typevars import T


def Identity(v: T) -> T:
    return v


def Consumer(_: T) -> NoReturn:
    """
    Consumes given value with no return
    """
    return None


def IdentityTrue(_: T) -> bool:
    """
    Returns True independently of given value
    """
    return True


def IdentityFalse(_: T) -> bool:
    """
    Returns False independently of given value
    """
    return False


def Sum(a: T, b: T) -> T:
    """
    Returns sum of two given value
    """
    return a + b


def PairSum(t: Tuple[T, T]) -> T:
    """
    Returns sum of first two values of Ordered Collection
    """
    return t[0] + t[1]


def First(t):
    """
    Returns first value of Ordered Collection
    """
    return t[0]


def Second(t):
    """
    Returns second value of Ordered Collection
    """
    return t[1]


def Multiply(a: T, b: T) -> T:
    """
    Returns product of two given values
    """
    return a * b


def Square(a: T) -> T:
    """
    Returns square of given value
    """
    return a * a


def OneArgRandom(_: T) -> float:
    """
    Returns random value independent of given value
    """
    return random()


def PairWith(*func: Callable[[T], T]):
    """Returns function that, when applied to value `e` returns `e, f1(f2(...fn(e)...))`"""

    def f(a):
        res = a
        for g in func:
            res = g(res)
        return a, res

    return f


def RPairWith(*func: Callable[[T], T]):
    """Returns function that, when applied to value `e` returns `f1(f2(...fn(e)...)), e`"""

    def f(a):
        res = a
        for g in func:
            res = g(res)
        return res, a

    return f


def TupleSum(t: Tuple) -> T:
    """Returns sum of given Ordered Collection"""
    if len(t) == 0:
        return 0
    tuple_sum = t[0]
    for i in range(1, len(t)):
        tuple_sum += t[i]
    return tuple_sum


def Compose(*func: Callable[[T], T]):
    """Returns function that, when applied to value `e` returns `f1(f2(...fn(e)...))` """

    def h(x):
        res = x
        for f in func:
            res = f(res)
        return res

    return h


from finq import FINQ


def test_join():
    a = [1, 2, 5, 7]
    expected = '1257'

    a_f = FINQ(a)

    assert a_f.map(str).join() == expected


def test_join_with_delim():
    a = [1, 2, 5, 7]
    expected = '1, 2, 5, 7'

    a_f = FINQ(a)

    assert a_f.map(str).join(', ') == expected



from finq import FINQ, PairSum, TupleSum


def test_cartesian_product():
    a = [1, 2, 5]
    b = [2, 3, 6]
    expected = [(1, 2), (1, 3), (1, 6), (2, 2), (2, 3), (2, 6), (5, 2), (5, 3), (5, 6)]

    assert FINQ(a).cartesian_product(b).to_list() == expected


def test_cartesian_product_with_mapping():
    a = [1, 2, 5]
    b = [2, 3, 6]
    expected = [3, 4, 7, 4, 5, 8, 7, 8, 11]

    assert FINQ(a).cartesian_product(b, PairSum).to_list() == expected


def test_cartesian_power():
    a = [1, 2, 5]
    expected = [(1, 1), (1, 2), (1, 5), (2, 1), (2, 2), (2, 5), (5, 1), (5, 2), (5, 5)]

    assert FINQ(a).cartesian_power(2).to_list() == expected


def test_cartesian_power_three():
    a = [1, 2]
    expected = [(1, 1, 1), (1, 1, 2), (1, 2, 1), (1, 2, 2), (2, 1, 1), (2, 1, 2), (2, 2, 1), (2, 2, 2)]

    assert FINQ(a).cartesian_power(3).to_list() == expected


def test_cartesian_first_power():
    a = [1, 2, 5]
    expected = [1, 2, 5]

    assert FINQ(a).cartesian_power(1).to_list() == expected


def test_cartesian_zero_power():
    a = [1, 2, 5]
    expected = []

    assert FINQ(a).cartesian_power(0).to_list() == expected


def test_cartesian_power_with_mapping():
    a = [1, 2, 5]
    expected = [2, 3, 6, 3, 4, 7, 6, 7, 10]

    assert FINQ(a).cartesian_power(2, PairSum).to_list() == expected


def test_cartesian_power_three_with_mapping():
    a = [1, 2]
    expected = [3, 4, 4, 5, 4, 5, 5, 6]

    assert FINQ(a).cartesian_power(3, TupleSum).to_list() == expected


from collections import defaultdict, Counter
from typing import Iterable, Generic, Set as TSet, List as TList, Dict as TDict, Counter as TCounter

from finq.constants import *
from finq.typevars import *


class FINQ(Iterable[T]):
    """Wrapper class for Iterables which simplifies processing of data"""

    def __init__(self, source: Iterable[T]):
        self._source = source

    def __iter__(self):
        for item in self._source:
            yield item

    def self(self, extension: Callable[['FINQ[T]'], TOut]):
        return extension(self)

    def concat(self, b: Iterable[T]) -> 'FINQ[T]':
        """Concatenates two sequences, creating sequence that contains items of the
         first iterable then of second iterable."""
        return FINQConcat(self, b)

    def map(self, func: Callable[[T], TOut]) -> 'FINQ[TOut]':
        """Applies given function to every element of sequence. """
        return FINQ(map(func, self))

    def zip(self, *b: TList[Iterable[T2]]) -> 'FINQ[Tuple[T2, ...]]':
        """Pairs corresponding elements of two sequences in pairs."""
        return FINQ(zip(self, *b))

    def flat_map(self, func: Callable[[T], Iterable[T2]] = Identity) -> 'FINQ[T2]':
        """Applies given function to every element to get collection, then glues these collections."""
        return FINQFlatMap(self, func)

    def flatten(self, flattener: Callable[[T], T2] = Identity) -> 'FINQ[T2]':
        """Applies given function to every element to get collection, then glues these collections.
        Repeats until all elements are non iterable."""
        return FINQFlatten(self, flattener)

    def filter(self, predicate: Callable[[T], T2]) -> 'FINQ[T]':
        """Removes elements that doesn't satisfy predicate from sequence."""
        return FINQ(filter(predicate, self))

    def distinct(self, func: Callable[[T], T2] = Identity) -> 'FINQ[T]':
        """Skips elements which `f(element)` repeated."""
        return FINQDistinct(self, func)

    def sort(self, func: Callable[[T], float] = Identity, /, reverse=False) -> 'FINQ[T2]':
        """Sorts sequence elements by key given by `f`."""
        return FINQ(sorted(self, key=func, reverse=reverse))

    def skip(self, count: int) -> 'FINQ[T]':
        """Skips `count` elements from sequence."""
        return FINQ(o for i, o in enumerate(self, 0) if i >= count)

    def take(self, count: int) -> 'FINQ[T]':
        """Limits sequence by `count` elements, dropping others."""
        return FINQ(o for i, o in enumerate(self, 0) if i < count)

    def cartesian_product(self, b: Iterable[T2], mapping: Callable[[Tuple], T] = None) -> 'FINQ[Tuple[T,T2]]':
        """Returns Cartesian product of two sequences after application of mapping if specified."""
        return FINQCartesianProduct(self, b, mapping)

    def cartesian_power(self, power: int, mapping: Callable[[Tuple], T] = None) -> 'FINQ':
        """Returns Cartesian power of sequence after application of mapping if specified."""
        return FINQCartesianPower(self, power, mapping)

    def pairs(self) -> 'FINQ[Tuple[T,T]]':
        """Returns Cartesian square of sequence. Equivalent to Cartesian square with Identity mapping."""
        return FINQPairs(self)

    def enumerate(self, start: int = 0) -> 'FINQ[Tuple[int, T]]':
        """Maps sequence elements to pair which first value is index in sequence starting by `start`."""
        return FINQ(enumerate(self, start))

    def peek(self, func: NoReturn = Identity) -> 'FINQ[T]':
        """Applies function to each element in sequence leaving sequence unchanged."""
        return FINQPeek(self, func)

    def group_by(self, func: Callable[[T], T2] = Identity) -> 'FINQ[TList[T]]':
        """Splits sequence into sequence of lists of elements which `f(x)` is the same."""
        return FINQGroupBy(self, func)

    def random(self, percentage: float) -> 'FINQ[T]':
        """Takes roughly `percentage*100%` of random elements of sequence."""
        return FINQ(i for i in self if random() < percentage)

    def shuffle(self) -> 'FINQ[T]':
        """Shuffles sequence."""
        return self.sort(OneArgRandom)

    def join_str(self, delimiter: str = '') -> str:
        """Joins sequence by `delimiter`."""
        return delimiter.join(self)

    def join(self, sequence: Iterable[T2], condition: Callable[[T, T2], bool],
             aggregate: Callable[[T, T2], T3]) -> 'FINQ[T3]':
        """Joins two sequences. Two values are aggregated if `condition` is true"""
        return FINQJoin(self, sequence, condition, aggregate)

    def for_each(self, func: NoReturn = Consumer) -> NoReturn:
        """Calls `f` for every element of a sequence."""
        for item in self:
            func(item)

    def all(self, func: Callable[[T], bool] = IdentityTrue) -> bool:
        """Checks if all elements in sequence satisfy predicate."""
        for i in self:
            if not func(i):
                return False
        return True

    def any(self, func: Callable[[T], bool] = IdentityTrue) -> bool:
        """Checks if there exist element in sequence that satisfies predicate."""
        for i in self:
            if func(i):
                return True
        return False

    def none(self, func: Callable[[T], bool] = IdentityTrue) -> bool:
        """Checks if there no element in sequence that satisfies predicate."""
        for i in self:
            if func(i):
                return False
        return True

    def first(self) -> T:
        """Takes first element of sequence."""
        return next(iter(self))

    def to_list(self) -> TList[T]:
        """Creates default python-list containing all sequence elements."""
        return list(self)

    def to_set(self) -> TSet[T]:
        """Creates default python-set containing all distinct sequence elements."""
        return set(self)

    def to_counter(self) -> TCounter[T]:
        """Creates Counter containing all sequence elements."""
        return Counter(self)

    def to_dict(self, key: Callable[[T], T1] = First, value: Callable[[T], T2] = Second) -> TDict[T1, T2]:
        """Creates default python-dict containing mapping `(key(x), value(x))` for every `x` in sequence."""
        if key == First and value == Second:
            return dict(self)
        return dict(self.map(lambda t: (key(t), value(t))))

    def count(self) -> int:
        """Returns count of elements in sequence."""
        return len(list(self))

    def min(self) -> T:
        """Finds minimal element in sequence."""
        return min(self)

    def max(self) -> T:
        """Finds maximal element in sequence."""
        return max(self)

    def sum(self) -> T:
        """Sums all elements of sequence. Works only for summable types."""
        return sum(self) or 0

    def max_diff(self) -> T:
        """Counts maximal difference between elements. Equal to difference between max and min for sequence."""
        max_v, min_v = None, None
        for i in self:
            if max_v is None or i > max_v:
                max_v = i
            if min_v is None or i < min_v:
                min_v = i
        return max_v - min_v

    def reduce(self, reducer: Callable[[T, T], T], /, first: T = None) -> T:
        """Applies function to first two elements, then to result and next element until elements end.
         Allows to specify first element."""
        return FINQReduce(self, reducer, first)

    def fold(self, mapper: Callable[[T], T2], aggregator: Callable[[T2, T2], T2], /) -> T2:
        """Applies mapper to each element, then aggregates pairs of T2 into single T2 until elements end.
        Equivalent to `finq.map(mapper).reduce(aggregator)`"""
        return FINQFold(self, mapper, aggregator)


class FINQFlatMap(FINQ[T]):
    """Applies given function to every element to get collection, then glues these collections."""

    def __init__(self, source: Iterable[T], mapper: Callable[[T], T2]):
        super().__init__(source)
        self.mapper = mapper

    def __iter__(self):
        for item in self._source:
            for sub_item in self.mapper(item):
                yield sub_item

    def self(self, extension: Callable[['FINQFlatMap[T]'], TOut]):
        return extension(self)


class FINQPairs(FINQ[T]):
    """Returns Cartesian square of sequence. Equivalent to Cartesian square with Identity mapping."""

    def __init__(self, source: Iterable[T]):
        super().__init__(source)

    def __iter__(self):
        src_list = list(self._source)
        for i in src_list:
            for item2 in src_list:
                yield i, item2

    def self(self, extension: Callable[['FINQPairs[T]'], TOut]):
        return extension(self)


class FINQPeek(FINQ[T]):
    """Applies function to each element in sequence leaving sequence unchanged."""

    def __init__(self, source: Iterable[T], func: NoReturn):
        super().__init__(source)
        self.func = func

    def __iter__(self):
        for item in self._source:
            self.func(item)
            yield item

    def self(self, extension: Callable[['FINQPeek[T]'], TOut]):
        return extension(self)


class FINQReduce(FINQ[T]):
    """Applies function to first two elements, then to result and next element until elements end.
    Allows to specify first element."""

    def __init__(self, source: Iterable[T], reducer: Callable[[T, T], T], first=None):
        super().__init__(source)
        self.reducer = reducer
        self.firstValue = first

    def __iter__(self):
        result = self.firstValue
        for item in self._source:
            if not result:
                result = item
                continue
            result = self.reducer(result, item)
        yield result

    def self(self, extension: Callable[['FINQReduce[T]'], TOut]):
        return extension(self)


class FINQFold(FINQ[T], Generic[T, T2]):
    """Applies mapper to each element, then aggregates pairs of T2 into single T2 until elements end.
    Allows to specify first element."""

    def __init__(self, source: Iterable[T], mapper: Callable[[T], T2], aggregator: Callable[[T2, T2], T2]):
        super().__init__(source)
        self.mapper = mapper
        self.aggregator = aggregator

    def __iter__(self):
        result = None
        for item in self._source:
            if not result:
                result = self.mapper(item)
                continue
            result = self.aggregator(result, self.mapper(item))
        yield result

    def self(self, extension: Callable[['FINQFold[T,T2]'], TOut]):
        return extension(self)


class FINQCartesianProduct(FINQ[T], Generic[T, T2]):
    """Returns Cartesian product of two sequences after application of mapping if specified."""

    def __init__(self, source: Iterable[T], b: Iterable[T2], mapping: Callable[[Tuple[T, T2]], T] = None):
        super().__init__(source)
        self.mapping = mapping
        self.b = b

    def __iter__(self) -> Iterable[Tuple[T, T2]]:
        b_list = list(self.b)
        if self.mapping is not None:
            for item in self._source:
                for b in b_list:
                    yield self.mapping((item, b))
        else:
            for item in self._source:
                for b in b_list:
                    yield item, b

    def self(self, extension: Callable[['FINQCartesianProduct[T, T2]'], TOut]):
        return extension(self)


class FINQCartesianPower(FINQ[T]):
    """Returns Cartesian power of sequence after application of mapping if specified."""

    def __init__(self, source: Iterable[T], power: int, mapping: Callable[[Tuple], T] = None):
        super().__init__(source)
        self.mapping = mapping
        self.power = power

    def __iter__(self):
        if self.power <= 0:
            return 0
        if self.power == 1:
            for k in self._source:
                yield k
            return
        items = list(self._source)
        for i in items:
            if self.mapping is not None:
                for j in FINQCartesianPower(items, self.power - 1):
                    if isinstance(j, tuple):
                        yield self.mapping((i, *j))
                    else:
                        yield self.mapping((i, j))
            else:
                for j in FINQCartesianPower(items, self.power - 1):
                    if isinstance(j, tuple):
                        yield i, *j
                    else:
                        yield i, j

    def self(self, extension: Callable[['FINQCartesianPower[T]'], TOut]):
        return extension(self)


class FINQGroupBy(FINQ[TList[T]]):
    """Splits sequence into sequence of lists of elements which `f(x)` is the same."""

    def __init__(self, source: Iterable[T], func: Callable[[T], T2]):
        super().__init__(source)
        self.func = func

    def __iter__(self):
        groups = defaultdict(list)
        for i in self._source:
            groups[self.func(i)].append(i)
        for k in groups.keys():
            yield groups[k]

    def self(self, extension: Callable[['FINQGroupBy[T]'], TOut]):
        return extension(self)


class FINQDistinct(FINQ[T]):
    """Skips elements which `f(element)` repeated."""

    def __init__(self, source: Iterable[T], func: Callable[[T], T2]):
        super().__init__(source)
        self.func = func

    def __iter__(self):
        looked = set()

        for item in self._source:
            var = self.func(item)
            if var not in looked:
                yield item
                looked.add(var)

    def self(self, extension: Callable[['FINQDistinct[T]'], TOut]):
        return extension(self)


class FINQConcat(FINQ[T]):
    """Concatenates two sequences, creating sequence that contains items of
    the first iterable then of second iterable."""

    def __init__(self, source: Iterable[T], b: Iterable[T]):
        super().__init__(source)
        self.b = b

    def __iter__(self):
        for i in self._source:
            yield i
        for i in self.b:
            yield i

    def self(self, extension: Callable[['FINQConcat[T]'], TOut]):
        return extension(self)


class FINQFlatten(FINQ[T]):
    """Applies given function to every element to get collection, then glues these collections.
    Repeats until all elements are non iterable."""

    def __init__(self, source: Iterable[T], flattener: Callable[[T], T2]):
        super().__init__(source)
        self.flattener = flattener

    def __iter__(self):
        for i in self._source:
            b = self.flattener(i)
            if isinstance(b, Iterable):
                yield from FINQFlatten(b, self.flattener)
            else:
                yield b

    def self(self, extension: Callable[['FINQFlatten[T]'], TOut]):
        return extension(self)


class FINQJoin(FINQ[T], Generic[T, T2, T3]):
    def __init__(self, source: Iterable[T], joined_sequence: Iterable[T2], condition: Callable[[T, T2], bool],
                 aggregator: Callable[[T, T2], T3]):
        super().__init__(source)
        self.aggregator = aggregator
        self.condition = condition
        self.joined_sequence = joined_sequence

    def __iter__(self):
        seq_list = list(self.joined_sequence)
        for i in self._source:
            for j in seq_list:
                if self.condition(i, j):
                    yield self.aggregator(i, j)

    def self(self, extension: Callable[['FINQJoin[T,T2, T3]'], TOut]):
        return extension(self)



from finq import FINQ


def test_first():
    a = [1, 2, 5, 7, 12, 15, 22, 26]

    a_f = FINQ(a)

    assert a_f.first() == 1


def test_first_twice():
    a = [1, 2, 5, 7, 12, 15, 22, 26]

    a_f = FINQ(a)

    assert a_f.first() == 1
    assert a_f.first() == 1


def test_to_set():
    a = [1, 2, 5, 7, 12, 15, 22, 26]
    expected = {1, 2, 5, 7, 12, 15, 22, 26}

    a_f = FINQ(a)

    assert a_f.to_set() == expected


def test_to_list():
    a = {1, 2, 5, 7, 12, 15, 22, 26}
    expected = [1, 2, 5, 7, 12, 15, 22, 26]

    a_f = FINQ(a)

    assert a_f.to_list() == expected



from finq import FINQ
from test.test_foreach import collector


def test_peek():
    a = [1, 2, 5, 7]
    expected = ['1', '1', '2', '2', '5', '5', '7', '7']

    mock = collector([])
    a_f = FINQ(a)

    a_f.map(str).peek(mock).for_each(mock)

    assert mock.collector == expected



from collections import Counter

from finq import FINQ


def test_group_by_identity():
    a = [1, 1, 2, 3, 3, 3]
    expected = [[1, 1], [2], [3, 3, 3]]

    assert FINQ(a).group_by().to_list() == expected


def test_group_by_func():
    a = [1, 1, 2, 3, 3, 3]
    expected = [[1, 1, 3, 3, 3], [2]]
    func = lambda i: i % 2

    assert FINQ(a).group_by(func).to_list() == expected


def test_flat_group_by_equals_argument():
    a = [1, 1, 2, 3, 3, 3]
    expected = Counter([1, 1, 2, 3, 3, 3])
    func = lambda i: i % 2

    assert FINQ(a).group_by(func).flatten().to_counter() == expected



from finq import FINQ


def test_reduce():
    a = [1, 2, 5, 7, 12, 15, 22, 26]
    expected = 90

    a_f = FINQ(a)

    assert a_f.reduce(lambda a, b: a + b) == expected


def test_reduce_with_first():
    a = [1, 2, 5, 7, 12, 15, 22, 26]
    expected = 100

    a_f = FINQ(a)

    assert a_f.reduce_with_first(10, lambda a, b: a + b) == expected



import setuptools

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read().replace('×', 'x').replace("²", "^2").replace("ⁿ", "^n")

setuptools.setup(
    name="FINQ",
    version="1.1.4b",
    author="FacelessLord",
    author_email="skyres21@gmail.com",
    description="Lightweight conveyor data processing python framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/FacelessLord/FINQ",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
)



from finq import FINQ


def test_int_map():
    a = [1, 2, 5, 7, 12, 15, 22, 26]
    expected = [1, 4, 25, 49, 144, 225, 484, 676]

    func = lambda i: i * i
    a_f = FINQ(a)

    assert a_f.map(func).to_list() == expected


def test_maps_str_to_int():
    a = ['1', '2', '5', '7', '12', '15', '22', '26']
    expected = [1, 2, 5, 7, 12, 15, 22, 26]

    a_f = FINQ(a)

    assert a_f.map(int).to_list() == expected


def inc_add(a):
    def inc(i):
        nonlocal a
        a += 1
        return i + a - 1

    return inc


def test_maps_with_captured_variable():
    a = [1, 2, 5, 7, 12, 15, 22, 26]
    expected = [2, 4, 8, 11, 17, 21, 29, 34]

    func = inc_add(1)

    a_f = FINQ(a)

    assert a_f.map(func).to_list() == expected
